// ==========================================================
// VanDerPol_RK
// CBP 27-01-23 
// Orig code FBO 19-01-23
// Allows passing of parameters down the chain
// ==========================================================

#include <Wire.h>
#include <Servo.h>

#define STATE_CALCULATE 1
#define STATE_TRANSMIT 2
#define STATE_READ 3

#define NEQ 2 //Largest required

#define DEBUG true

#define HEAD 1
#define SEGM 2
#define TAIL 3

#define LED_PIN 3

int segType;

int this_node_ID;
int target_node_ID;

int nrData;           // nr of data elements in a package

int state;
bool endFound = false;

char message[20];
int i;

float u[NEQ];
float dudt[NEQ];
float yout[NEQ];

float deltaT;
float time;

float input;

// parameters for head, seg and tail
float eps;
float k;

// Scaling of PWD for LED
float minVal;
float maxVal;

void setup() {

  pinMode(LED_PIN, OUTPUT);
  Serial.begin(9600);

  deltaT = 0.01;
  time = 0;

  // Model parameters -----------------------------------------------
  // TAIL and SEGM
  eps = 0.1;
  k = 0.1;

  // Configuration --------------------------------------------------

  nrData = 5;
  // First float sent is eps
  // Second float sent is k
  // Third float sent is u[0]
  // Fourth float is minVal
  // Fifth float is maxVal

  // min and max vals for pwm scaling
  minVal = -3;
  maxVal =  3;

  //segType = HEAD;
  //segType = SEGM;
  segType = TAIL;

  if (segType == HEAD){
    this_node_ID = 0;
    target_node_ID = 1;
    state = STATE_CALCULATE;
  }

  if(segType == SEGM){
    this_node_ID = 1;
    target_node_ID = this_node_ID + 1;
    state = STATE_READ;
  }

  if(segType == TAIL){
    this_node_ID = 1;
    target_node_ID = 0;
   state = STATE_READ;
  }

  // Calc Initial Conditions ----------------------------------------

  if(segType == HEAD){
    u[0] = (float)random(0,10000)/10000.0;
    u[1] = (float)random(0,10000)/10000.0;
  }
  if(segType == SEGM || segType == TAIL){
    u[0] = (float)random(0,10000)/10000.0;
    u[1] = (float)random(0,10000)/10000.0;
  }

  // ----------------------------------------------------------------

  Serial.print("segType = ");
  Serial.println(segType);
  Serial.print("This node = ");
  Serial.println(this_node_ID);
  Serial.print("Target node = ");
  Serial.println(target_node_ID);

  Wire.begin(this_node_ID);
  Wire.onReceive(receiveEvent);

}

void loop() {

  float op;
  if (state == STATE_CALCULATE) {
    if(DEBUG) Serial.println("STATE_CALCULATE");
    solver_RK(NEQ, u, dudt, time, deltaT, yout, *rhs_segment);
    time += deltaT;
    for (i = 0; i < NEQ; i++) {
      u[i] = yout[i];
    }

    op = u[0];
    //scaling for the PWM
    int pwm = 255*(op - minVal)/(maxVal - minVal);
    if(pwm < 0) pwm = 0;
    if(pwm>255) pwm = 255;
    analogWrite(LED_PIN, pwm);
    if(DEBUG) Serial.print("u[0] = ");
    if(DEBUG) Serial.println(u[0]);
    if(DEBUG) Serial.print("  op = ");
    if(DEBUG) Serial.println(op);
    if(DEBUG) Serial.print("  pwm = ");
    if(DEBUG) Serial.println(pwm);
    //Serial.println(message);
    state = STATE_TRANSMIT;
  }

  if (state == STATE_TRANSMIT) {
    if (DEBUG) Serial.print("STATE_TRANSMIT: op = ");
    if (DEBUG) Serial.print(op);
    delay(200);
    Wire.beginTransmission(target_node_ID);
    Wire.print('H');
    Wire.print(nrData);
    Wire.print(',');
    Wire.print(eps);
    Wire.print(',');
    Wire.print(k);
    Wire.print(',');
    Wire.print(minVal);
    Wire.print(',');
    Wire.print(maxVal);
    Wire.print(',');
    Wire.print(op);
    Wire.println();
    byte status = Wire.endTransmission();
    if(status != 0x00) {
      Serial.println("Bus error");
    }
    delay(100);

    state = STATE_READ;
  }

  if (state == STATE_READ) {
    if (endFound) {
      state = STATE_CALCULATE;
      endFound = false;
    }
  }

}

void receiveEvent(int howMany) {
  if (DEBUG) Serial.println("receiveEvent from ");
  //Serial.println(target_node_ID);
  endFound = false;
  int i = 0;
  while (Wire.available() > 0) {
    char c = Wire.read(); // receive byte as a character
    if (DEBUG) Serial.println(c);
    if (c == 'H') {
      // first get nrData items
      int nrDataItems = Wire.parseInt();
      if (DEBUG) Serial.println(nrDataItems);      
      float float1 = Wire.parseFloat();
      eps = float1;
      float float2 = Wire.parseFloat();
      k = float2;
      float float3 = Wire.parseFloat();
      minVal = float3;
      float float4 = Wire.parseFloat();
      maxVal = float4;
      float float5 = Wire.parseFloat();
      input = float5;            
      if (DEBUG) { Serial.print("eps = "); Serial.println(float1); }
      if (DEBUG) { Serial.print("k = ");   Serial.println(float2); }
      if (DEBUG) { Serial.print("val = "); Serial.println(float3); }
      endFound = true;
    }
  }
}


void rhs_segment(float t, float u[], float dudt[]) {

  if(segType == HEAD){
      dudt[0] = u[1];
      dudt[1] = eps*(1-u[0]*u[0])*u[1] - u[0] + k*(u[0] - input);
  }

  if(segType == SEGM || segType == TAIL){
      dudt[0] = u[1];
      dudt[1] = eps*(1-u[0]*u[0])*u[1] - u[0] + k*(u[0] - input);
  }

}

void solver_RK(int nVar, float y[], float* dydx, float t, float h, float* yout, void (*derivs) (float, float*, float*)) {
  float hh = h * 0.5;
  float h6 = h / 6.0;
  float xh = t + hh;
  float yt[NEQ];
  float dym[NEQ];
  float dyt[NEQ];

  derivs(t,y,dydx);

  for (int i = 0; i < nVar; i++) {
    yt[i] = y[i] + hh * dydx[i];
  }
  derivs(xh, yt, dyt);

  for (int i = 0; i < nVar; i++) {
    yt[i] = y[i] + hh * dyt[i];
  }
  derivs(xh, yt, dym);

  for (int i = 0; i < nVar; i++) {
    yt[i] = y[i] + h * dym[i];
    dym[i] += dyt[i];
  }
  derivs(t + h, yt, dyt);

  for (int i = 0; i < nVar; i++) {
    yout[i] = y[i] + h6 * (dydx[i] + dyt[i] + 2.0 * dym[i]);
  }
}
